"""
Droplet Contact Numbers and Probabilities (DCNumPro)

Determine droplet contact numbers and probabilities based on 
the median and the standard deviation of a droplet size distribution.

Source code of DCNumPro: https://git.rwth-aachen.de/avt-fvt/public/contact-numbers-and-probabilities/
DCNumPro uses a 3D cell-based Voronoi library based on voro++: https://github.com/wackywendell/tess

The source code of DCNumPro is licensed under the Eclipse Public License v2.0 (EPL 2.0).
Copyright (c) 2024 Fluid Process Engineering (AVT.FVT), RWTH Aachen University.
Written by Stepan Sibirtsev & Andrey Kirsanov

The coyprights and license terms are given in LICENSE.

This script contains all functions to determine contact numbers and probabilities.
Script version date: 23.11.2024
"""  

""" Import packages """
import numpy as np
from collections import defaultdict
from collections import Counter

""" Functions """
def calc_contact_probabilities(average_contacts, contact_probabilities, cell_face_fixed, diameters_neighbors, diameters_0, grow_step, particle_status):
    """ 
    This function calculates the average contact number of each class and the contact probability between classes.
    Particles which are in contact with walls at their initial state are excluded.
    If solver issues occur at growing simulation of particles those particles are excluded.
    Wall contacts are excluded.
    Only contacts between fixed faces are included in the calculation.    
    """
    """ Count contacts """
    # Initialize a dictionary to hold the contact counts
    contact_counts = defaultdict(lambda: defaultdict(int))
    # Iterate over each particle and its neighbors to count the contacts
    for particle_i in range(len(diameters_0)):
        particle = diameters_0[particle_i]
        for neighbor_j in range(len(diameters_neighbors[particle_i])):
            neighbor = diameters_neighbors[particle_i][neighbor_j]
            # Exclude particle with status "False", wall contacts (None values) and faces which are not fixed
            if particle_status[particle_i] != False and neighbor != None and cell_face_fixed[particle_i][neighbor_j] == True:  
                # Increase number of contacts
                contact_counts[particle][neighbor] += 1
    """ Calculate total contacts """
    # Initialize a dictionary to hold the total contacts for each particle class
    total_contacts = defaultdict(int)
    # Sum the total contacts for each particle class
    for class_particle in contact_counts:
        total_contacts[class_particle] += sum(contact_counts[class_particle].values())
    # Sort value_counts by keys (values) in ascending order and convert to a list of tuples
    total_contacts_sorted = sorted(total_contacts.items())
    # Extract only the counts (second column)
    total_contacts_values = [count for value, count in total_contacts_sorted]
    """ Calculate class sizes """
    # Count occurrences of each diameter 
    class_size = Counter()
    # Iterate through data_list and include values based on include_list
    for class_diameter, status in zip(diameters_0, particle_status):
        if status == True:
            class_size[class_diameter] += 1
    # Sort value_counts by keys (values) in ascending order and extract only the counts
    class_size_values = [class_size[value] for value in sorted(class_size)]
    """ Calculate average contact nubmers based on class sizes and total contacts """
    # Perform element-wise division
    for i in range(len(class_size_values)):
        average_contacts[grow_step][i] = total_contacts_values[i] / class_size_values[i]
    """ Calculate contact probability based on the contact counts and total contacts """
    # Initialize a dictionary to hold the probabilities
    contact_probabilities_dict = defaultdict(lambda: defaultdict(float))
    # Iterate over each particle and its neighbors to calculate the probabilities
    for class_particle, classes_neighbors in contact_counts.items():
        for class_neighbor, count_contacts in classes_neighbors.items():
            # Avoid division by zero
            if total_contacts[class_particle] > 0:  
                # Calculate the contact probabilities
                contact_probabilities_dict[class_particle][class_neighbor] = 100 * count_contacts / total_contacts[class_particle]
    # Convert the contact_probabilities dictionary into a sorted list (sorted in ascending order)
    contact_counts_sorted = sorted(set(contact_counts))
    # Iterate over each particle and its neighbors to update the list in ascending order
    for class_particle_i in range(len(contact_counts_sorted)):
        class_particle = contact_counts_sorted[class_particle_i]
        row = []
        for class_neighbor in contact_counts_sorted:
            row.append(contact_probabilities_dict[class_particle].get(class_neighbor, 0))
        contact_probabilities[grow_step][class_particle_i] = row

    return average_contacts, contact_probabilities